/*
Copyright (C) 2001 Deflection

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <windows.h>
#define DIRECTINPUT_VERSION 0x0700
#include <dinput.h>
#include "Controller.h"
#include "DI.h"
#include "DefDI.h"
#include "Config.h"
#include "resource.h"


HINSTANCE g_hInstance;

GUID Guids[MAX_DEVICES];
DEFCONTROLLER Controller[NUMBER_OF_CONTROLS];
CONTROL *ControlDef[4];

int WINAPI DllMain ( HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved) {
    
	switch (fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			g_hInstance = hInstance;
			break;

		case DLL_PROCESS_DETACH:
			FreeDirectInput();
			break;
	}

	return TRUE;
}

EXPORT void CALL CloseDLL (void) {
	//Stop and Close Direct Input
	FreeDirectInput();
}

EXPORT void CALL ControllerCommand ( int Control, BYTE * Command) {
}

EXPORT void CALL DllAbout ( HWND hParent ) {
	MessageBox(hParent,"Def's Direct Input 0.54\nFor DirectX 7 or higher\n\nby Deflection","About",MB_ICONINFORMATION | MB_OK);
}

EXPORT void CALL DllConfig ( HWND hParent ) {
	if (g_lpDI == NULL)
		InitializeAndCheckDevices(hParent);
	else
		CreateDialog ( g_hInstance, MAKEINTRESOURCE(IDD_CONFIGDLG), hParent, (DLGPROC)ConfigDlgProc);
}

EXPORT void CALL GetDllInfo ( PLUGIN_INFO * PluginInfo ) {
	PluginInfo->Version = 0x0100;
	PluginInfo->Type = PLUGIN_TYPE_CONTROLLER;
	wsprintf(PluginInfo->Name,"Def's Direct Input 0.54");
}

EXPORT void CALL GetKeys(int Control, BUTTONS * Keys ) {
	BUTTONS	ControllerInput;
	
	//Empty Keyboard Button Info 
	ControllerInput.Value = 0;
	
	if (Controller[Control].bActive == TRUE) {
		BYTE     buffer[256];   //Keyboard Info 
		DIJOYSTATE js;          //Joystick Info
		HRESULT  hr;
		int M1Speed = 0, M2Speed = 0;

		if (Keys == NULL) { return; }

		for(BYTE devicecount=0; devicecount<Controller[Control].NDevices; devicecount++) {		
			BYTE DeviceNum = (BYTE) Controller[Control].Devices[devicecount];
			if (DInputDev[DeviceNum].lpDIDevice != NULL) {
				LONG count;
		
				if((DInputDev[DeviceNum].DIDevInst.dwDevType & DIDEVTYPE_KEYBOARD) == DIDEVTYPE_KEYBOARD){			
					ZeroMemory( &buffer, sizeof(buffer) );	
					if FAILED(hr = DInputDev[DeviceNum].lpDIDevice->GetDeviceState(sizeof(buffer),&buffer)) { 
						hr = DInputDev[DeviceNum].lpDIDevice->Acquire();
							while( hr == DIERR_INPUTLOST ) 
								hr = DInputDev[DeviceNum].lpDIDevice->Acquire();
						return;
					}
						
					for (count=0;count<NUMBER_OF_BUTTONS;count++) {
						if ( Controller[Control].Input[count].Device == DeviceNum ) {
							switch ( Controller[Control].Input[count].type ) {
								//Record Keyboard Button Info from Device State Buffer
								case INPUT_TYPE_KEY_BUT:
									if (BUTTONDOWN(buffer, Controller[Control].Input[count].vkey)) {
										switch (count)
										{
											case 18:
												M1Speed = Controller[Control].Input[count].button; 
												break;

											case 19:
												M2Speed = Controller[Control].Input[count].button;
												break;

											default:
												ControllerInput.Value |= Controller[Control].Input[count].button;
												break;
										}
									}
									break;

								default:
									break;
							}
						}
					}
				}

				else if((DInputDev[DeviceNum].DIDevInst.dwDevType & DIDEVTYPE_JOYSTICK) == DIDEVTYPE_JOYSTICK){
					if FAILED(hr = DInputDev[DeviceNum].lpDIDevice->Poll()) { 
						hr = DInputDev[DeviceNum].lpDIDevice->Acquire();
						while( hr == DIERR_INPUTLOST ) 
							hr = DInputDev[DeviceNum].lpDIDevice->Acquire();
						return; 
					}
					if FAILED( hr = DInputDev[DeviceNum].lpDIDevice->GetDeviceState( sizeof(DIJOYSTATE), &js ) )
						return;
				    
					for (count=0;count<NUMBER_OF_BUTTONS;count++) {
						if ( Controller[Control].Input[count].Device == DeviceNum ) {
							BYTE count2;
							switch ( Controller[Control].Input[count].type ) 
							{
								//Get Joystick button Info from Device State js stucture
								case INPUT_TYPE_JOY_BUT:
									if (BUTTONDOWN(js.rgbButtons, Controller[Control].Input[count].vkey)) {
										switch (count)
										{
											case 18:
												M1Speed = Controller[Control].Input[count].button; 
												break;

											case 19:
												M2Speed = Controller[Control].Input[count].button;
												break;

											default:
												ControllerInput.Value |= Controller[Control].Input[count].button;
												break;
										}
									}
									break;
						
								case INPUT_TYPE_JOY_AXIS:
									switch (Controller[Control].Input[count].vkey)
									{
										case DIJOFS_YN:
											if( js.lY < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lY, Control, count, &ControllerInput, M1Speed, M2Speed);
											} 
											break;
										case DIJOFS_YP:
											if( js.lY > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lY, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_XN:
											if( js.lX < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lX, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_XP:
											if( js.lX > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lX, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_ZN:
											if( js.lZ < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lZ, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_ZP:
											if( js.lZ > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lZ, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RYN:
											if( js.lRy < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lRy, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RYP:
											if( js.lRy > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lRy, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RXN:
											if( js.lRx < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lRx, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RXP:
											if( js.lRx > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lRx, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RZN:
											if( js.lRz < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.lRz, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_RZP:
											if( js.lRz > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.lRz, Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_SLIDER0N:
											if( js.rglSlider[0] < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.rglSlider[0], Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_SLIDER0P:
											if( js.rglSlider[0] > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.rglSlider[0], Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_SLIDER1N:
											if( js.rglSlider[1] < (LONG) -Controller[Control].SensMin ) {
												GetNegAxisVal(js.rglSlider[1], Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
										case DIJOFS_SLIDER1P:
											if( js.rglSlider[1] > (LONG) Controller[Control].SensMin ) {
												GetPosAxisVal(js.rglSlider[1], Control, count, &ControllerInput, M1Speed, M2Speed);
											}
											break;
									}
									break;
								
								case INPUT_TYPE_JOY_POV:
									for (count2=0;count2<4;count2++) {
										if ( (js.rgdwPOV[count2] != -1)  && (LOWORD(js.rgdwPOV[count2]) != 0xFFFF) ) {
											switch(Controller[Control].Input[count].vkey)
											{
												case DIJOFS_POV0N:
												case DIJOFS_POV1N:
												case DIJOFS_POV2N:
												case DIJOFS_POV3N:
													if( (js.rgdwPOV[count2] >= 31500) || (js.rgdwPOV[count2] <= 4500) ) {
														switch (count)
														{
															case 18:
																M1Speed = Controller[Control].Input[count].button; 
																break;

															case 19:
																M2Speed = Controller[Control].Input[count].button;
																break;

															default:
																ControllerInput.Value |= Controller[Control].Input[count].button;
																break;
														}

													}
													break;
												case DIJOFS_POV0E:
												case DIJOFS_POV1E:
												case DIJOFS_POV2E:
												case DIJOFS_POV3E:
													if ( (js.rgdwPOV[count2] >= 4500) && (js.rgdwPOV[count2] <= 13500) ) {
														switch (count2)
														{
															case 18:
																M1Speed = Controller[Control].Input[count].button; 
																break;

															case 19:
																M2Speed = Controller[Control].Input[count].button;
																break;

															default:
																ControllerInput.Value |= Controller[Control].Input[count].button;
																break;
														}
													}
													break;
												case DIJOFS_POV0S:
												case DIJOFS_POV1S:
												case DIJOFS_POV2S:
												case DIJOFS_POV3S:
													if ( (js.rgdwPOV[count2] >= 13500) && (js.rgdwPOV[count2] <= 22500) ) {
														switch (count2)
														{
															case 18:
																M1Speed = Controller[Control].Input[count].button; 
																break;

															case 19:
																M2Speed = Controller[Control].Input[count].button;
																break;

															default:
																ControllerInput.Value |= Controller[Control].Input[count].button;
																break;
														}
													}
													break;
												case DIJOFS_POV0W:
												case DIJOFS_POV1W:
												case DIJOFS_POV2W:
												case DIJOFS_POV3W:
													if ( (js.rgdwPOV[count2] >= 22500) && (js.rgdwPOV[count2] <= 31500) ) {
														switch (count2)
														{
															case 18:
																M1Speed = Controller[Control].Input[count].button; 
																break;

															case 19:
																M2Speed = Controller[Control].Input[count].button;
																break;

															default:
																ControllerInput.Value |= Controller[Control].Input[count].button;
																break;
														}
													}
													break;
											}
										}
									}

								default:
									break;
							}
						}
					}
				}
			}
		}
		
		if(M2Speed) {
			if ( ControllerInput.Y_AXIS < 0 ) 
				ControllerInput.Y_AXIS = (char) -M2Speed;
			else if ( ControllerInput.Y_AXIS > 0 )
				ControllerInput.Y_AXIS = (char) M2Speed;
			
			if ( ControllerInput.X_AXIS < 0 )
				ControllerInput.X_AXIS = (char) -M2Speed;
			else if ( ControllerInput.X_AXIS > 0 )
				ControllerInput.X_AXIS = (char) M2Speed;
		}
		if(M1Speed) {
			if ( ControllerInput.Y_AXIS < 0 ) 
				ControllerInput.Y_AXIS = (char) -M1Speed;
			else if ( ControllerInput.Y_AXIS > 0 )
				ControllerInput.Y_AXIS = (char) M1Speed;
			
			if ( ControllerInput.X_AXIS < 0 )
				ControllerInput.X_AXIS = (char) -M1Speed;
			else if ( ControllerInput.X_AXIS > 0 )
				ControllerInput.X_AXIS = (char) M1Speed;
		}
	}

	//Pass Button Info to Emulator
	Keys->Value = ControllerInput.Value;
	
}


void WINAPI GetNegAxisVal(LONG AxisValue, int Control, LONG count, BUTTONS *ControllerInput, int &M1Speed, int &M2Speed) {

	switch(count)
	{
		case 0:
			if ( AxisValue < (LONG) -Controller[Control].SensMax)
				ControllerInput->Y_AXIS = Controller[Control].SensMax;
			else
				ControllerInput->Y_AXIS = -AxisValue; 
			break;
		case 1:
			if ( AxisValue < (LONG) -Controller[Control].SensMax)
				ControllerInput->Y_AXIS = -Controller[Control].SensMax;
			else
				ControllerInput->Y_AXIS = AxisValue;
			break;
		case 2:
			if ( AxisValue < (LONG) -Controller[Control].SensMax)
				ControllerInput->X_AXIS = -Controller[Control].SensMax;
			else
				ControllerInput->X_AXIS = AxisValue; 
			break;
		case 3:
			if ( AxisValue < (LONG) -Controller[Control].SensMax)
				ControllerInput->X_AXIS = Controller[Control].SensMax;
			else
				ControllerInput->X_AXIS = -AxisValue; 
			break;
		
		case 18:
			M1Speed = Controller[Control].Input[count].button; 
			break;
		case 19:
			M2Speed = Controller[Control].Input[count].button;
			break;

		default: 
			ControllerInput->Value |= Controller[Control].Input[count].button; 
			break;
	}
}

void WINAPI GetPosAxisVal(LONG AxisValue, int Control, LONG count, BUTTONS *ControllerInput, int &M1Speed, int &M2Speed) {
	switch(count)
	{
		case 0:
			if ( AxisValue > (LONG) Controller[Control].SensMax)
				ControllerInput->Y_AXIS = Controller[Control].SensMax;
			else
				ControllerInput->Y_AXIS = AxisValue;
			break;
		case 1:
			if ( AxisValue > (LONG) Controller[Control].SensMax)
				ControllerInput->Y_AXIS = -Controller[Control].SensMax;
			else
				ControllerInput->Y_AXIS = -AxisValue; 
			break;
		case 2:
			if ( AxisValue > (LONG) Controller[Control].SensMax)
				ControllerInput->X_AXIS = -Controller[Control].SensMax;
			else
				ControllerInput->X_AXIS = -AxisValue; 
			break;
		case 3:
			if ( AxisValue > (LONG) Controller[Control].SensMax)
				ControllerInput->X_AXIS = Controller[Control].SensMax;
			else
				ControllerInput->X_AXIS = AxisValue; 
			break;
		
		case 18:
			M1Speed = Controller[Control].Input[count].button; 
			break;
		case 19:
			M2Speed = Controller[Control].Input[count].button;
			break;

		default: 
			ControllerInput->Value |= Controller[Control].Input[count].button; 
			break;
	}
}


EXPORT void CALL InitiateControllers (HWND hMainWindow, CONTROL Controls[4]) {
	HKEY hKey;
	TCHAR subkey[] = TEXT("Software\\DefRow\\Def's DI Plugin");
	DWORD dwSize, dwType;

	for (BYTE i=0; i<NUMBER_OF_CONTROLS; i++) {
		ControlDef[i] = &Controls[i];
		ControlDef[i]->Present = FALSE;
		ControlDef[i]->RawData = FALSE;
		ControlDef[i]->Plugin  = PLUGIN_NONE;

		Controller[i].NDevices = 0;
		Controller[i].bActive = FALSE;
		Controller[i].SensMax = 80;
		Controller[i].SensMin = 30;
		Controller[i].Input[18].button = 40;
		Controller[i].Input[19].button = 20;
		wsprintf(Controller[i].szName,"Controller %d",i+1);		
	}

	if (g_lpDI == NULL) {
		InitializeAndCheckDevices(hMainWindow);
	}
	
	dwType = REG_BINARY;
	dwSize = sizeof(DEFCONTROLLER);

	if ( RegCreateKeyEx(HKEY_CURRENT_USER, subkey, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hKey, 0) != ERROR_SUCCESS ) {
		MessageBox(NULL, "Could not create Registry Key", "Error", MB_ICONERROR | MB_OK);
	}
	else {
		for (BYTE NController = 0; NController<NUMBER_OF_CONTROLS; NController++) {
			if ( RegQueryValueEx(hKey, Controller[NController].szName, 0, &dwType, (LPBYTE)&Controller[NController], &dwSize) == ERROR_SUCCESS ) {
				if ( Controller[NController].bActive )
					ControlDef[NController]->Present = TRUE;
				else
					ControlDef[NController]->Present = FALSE;
				
				if ( dwSize != sizeof(DEFCONTROLLER) ) {
					dwType = REG_BINARY;
					dwSize = sizeof(DEFCONTROLLER);
					ZeroMemory( &Controller[NController], sizeof(DEFCONTROLLER) );
					
					Controller[NController].NDevices = 0;
					Controller[NController].bActive = FALSE;
					ControlDef[NController]->Present = FALSE;
					Controller[NController].SensMax = 80;
					Controller[NController].SensMin = 30;
					Controller[NController].Input[18].button = 40;
					Controller[NController].Input[19].button = 20;
					wsprintf(Controller[NController].szName,"Controller %d",NController+1);

					RegDeleteValue(hKey, Controller[NController].szName);
					RegSetValueEx(hKey, Controller[NController].szName, 0, dwType, (LPBYTE)&Controller[NController], dwSize);
				}
			}
			else {
				dwType = REG_BINARY;
				dwSize = sizeof(DEFCONTROLLER);
				RegDeleteValue(hKey, Controller[NController].szName);
				RegSetValueEx(hKey, Controller[NController].szName, 0, dwType, (LPBYTE)&Controller[NController], dwSize);
			}
		}		
	}
	RegCloseKey(hKey);
}

void WINAPI InitializeAndCheckDevices(HWND hMainWindow) {
	HKEY hKey;
	TCHAR subkey[] = TEXT("Software\\DefRow\\Def's DI Plugin");
	BYTE i;
	DWORD dwSize, dwType;
	
	//Initialize Direct Input function
	if( FAILED(InitDirectInput(hMainWindow)) ) {
		MessageBox(NULL,"DirectInput Initialization Failed!","Error",MB_ICONERROR | MB_OK);
		FreeDirectInput();
	}
	else {
		dwType = REG_BINARY;
		dwSize = sizeof(Guids);
		//Check Guids for Device Changes
		RegCreateKeyEx(HKEY_CURRENT_USER, subkey, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hKey, 0);
		if( RegQueryValueEx(hKey, TEXT("Guids"), 0, &dwType, (LPBYTE)Guids, &dwSize) != ERROR_SUCCESS ) {
			for(i=0; i<MAX_DEVICES; i++) {
				if ( DInputDev[i].lpDIDevice == NULL )
					ZeroMemory(&Guids[i], sizeof(GUID));
				else	
					memcpy(&Guids[i], &DInputDev[i].DIDevInst.guidInstance, sizeof(GUID));
			}
			dwType = REG_BINARY;
			RegSetValueEx(hKey, TEXT("Guids"), 0, dwType, (LPBYTE)Guids, dwSize);
		}
		else {
			if( CheckForDeviceChange(hKey) ) {
				for(i=0; i<MAX_DEVICES; i++) {
					if ( DInputDev[i].lpDIDevice == NULL )
						ZeroMemory(&Guids[i], sizeof(GUID));
					else	
						memcpy(&Guids[i], &DInputDev[i].DIDevInst.guidInstance, sizeof(GUID));
				}
				dwType = REG_BINARY;
				RegSetValueEx(hKey, TEXT("Guids"), 0, dwType, (LPBYTE)Guids, dwSize);
			}
		}
		RegCloseKey(hKey);
	}
}

BOOL WINAPI CheckForDeviceChange(HKEY hKey) {
	BOOL DeviceChanged;
	DWORD dwSize, dwType;
	
	dwType = REG_BINARY;
	dwSize = sizeof(DEFCONTROLLER);

	DeviceChanged = FALSE;
	
	for(BYTE DeviceNumCheck=0; DeviceNumCheck<MAX_DEVICES; DeviceNumCheck++) {
		if( memcmp(&Guids[DeviceNumCheck], &DInputDev[DeviceNumCheck].DIDevInst.guidInstance, sizeof(GUID)) != 0) {
			DeviceChanged = TRUE;			
			for( BYTE NController=0; NController<NUMBER_OF_CONTROLS; NController++) {
				RegQueryValueEx(hKey, Controller[NController].szName, 0, &dwType, (LPBYTE)&Controller[NController], &dwSize);
				for( BYTE DeviceNum=0; DeviceNum<Controller[NController].NDevices; DeviceNum++) {
					if( Controller[NController].Devices[DeviceNum] == DeviceNumCheck) {
						Controller[NController].NDevices = 0; 
						Controller[NController].bActive = FALSE;
						RegSetValueEx(hKey, Controller[NController].szName, 0, dwType, (LPBYTE)&Controller[NController], dwSize);
					}
				}
			}
		}
	}

	return DeviceChanged;
}

EXPORT void CALL ReadController ( int Control, BYTE * Command ) {
}

EXPORT void CALL RomClosed (void) {
}

EXPORT void CALL RomOpen (void) {
	HKEY hKey;
	TCHAR subkey[] = TEXT("Software\\DefRow\\Def's DI Plugin");
	DWORD dwSize, dwType;

	dwType = REG_BINARY;
	dwSize = sizeof(DEFCONTROLLER);
	
	if (RegOpenKeyEx(HKEY_CURRENT_USER, subkey, 0, KEY_READ, &hKey) == ERROR_SUCCESS ) {
		for ( BYTE NControl = 0; NControl<NUMBER_OF_CONTROLS; NControl++ ) {
			RegQueryValueEx(hKey, Controller[NControl].szName, 0, &dwType, (LPBYTE)&Controller[NControl], &dwSize);
			if ( Controller[NControl].bActive )
				ControlDef[NControl]->Present = TRUE;
			else
				ControlDef[NControl]->Present = FALSE;

		}
	}
	RegCloseKey(hKey);
}

EXPORT void CALL WM_KeyDown( WPARAM wParam, LPARAM lParam ) {
}

EXPORT void CALL WM_KeyUp( WPARAM wParam, LPARAM lParam ) {
}











